using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using TMPro;
using UnityEngine;
using UnityEngine.Events;

public class GrapplingGunBehaviour : MonoBehaviour, Messages.IUse
{   
    #region Objs
    private Rigidbody2D rb;   
    private PhysicalBehaviour phys;
    private SpriteRenderer sr;
    private LineRenderer lr;
    private SpawnableAsset projectileAsset = ModAPI.FindSpawnable("Hook - TGGM");
    public GameObject hookRef;
    #endregion

    #region Winch
    private DistanceJoint2D winch;
    #endregion

    #region Vectors
    public Vector2 barrelPosition;
    public Vector2 barrelDirection;
    public Vector2 GetBarrelPosition() => (Vector2) this.transform.TransformPoint((Vector3) this.barrelPosition);
    public Vector2 GetBarrelDirection() => (Vector2) (this.transform.TransformDirection((Vector3) this.barrelDirection) * this.transform.localScale.x);
    #endregion

    #region Vars
    public UnityEngine.Sprite[] text = new UnityEngine.Sprite[] {tggm.main.Assets.GrapplingGunLoaded, tggm.main.Assets.GrapplingGunUnloaded};
    private readonly HashSet<UnityEngine.Object> linkedObjects = new HashSet<UnityEngine.Object>();
    private WinchStateEnum winchState;
    private bool launched;
    private float maxDistance = 5000f;
    private float minDistance = 0.15f;
    public float projectileLaunchStrength = 2f;
    public float recoilMultiplier = 0.2f;
    #endregion 

    #region Setup
    private void Awake()
    {
        rb = GetComponent<Rigidbody2D>();
        phys = GetComponent<PhysicalBehaviour>();
        sr = GetComponent<SpriteRenderer>();

        sr.sprite = text[0];

        hookRef = (GameObject) null; 
        winchState = GrapplingGunBehaviour.WinchStateEnum.Idle;
    } 
    private void Start(){
        lr = gameObject.AddComponent<LineRenderer>();

        lr.material = ModAPI.FindMaterial("Sprites-Default");
        lr.startColor = lr.endColor = Color.black;
        lr.startWidth = 0.05f;
        lr.numCapVertices = 10;
        lr.numCornerVertices = 10;
        lr.positionCount = 0;

        AddContextMenu();    
    }
    private void AddContextMenu()
    {

        List<ContextMenuButton> buttons = GetComponent<PhysicalBehaviour>().ContextMenuOptions.Buttons;
        buttons.Add(new ContextMenuButton("Stop Winch", "Stop", "Stop the grappling gun's winch", () =>
        {
            ResetWinchVars();
        }));
        buttons.Add(new ContextMenuButton("retractWinch", "Retract", "Retract the grappling gun's winch", () =>
        {
            ResetWinchVars();
            winchState = GrapplingGunBehaviour.WinchStateEnum.Retracting;
        }));
        buttons.Add(new ContextMenuButton("extendWinch", "Extend", "Extend the grappling gun's winch", () =>
        {
            ResetWinchVars();
            winchState = GrapplingGunBehaviour.WinchStateEnum.Extending;
        }));
        buttons.Add(new ContextMenuButton("setMaxLength", "Set rope length", "Set the length of the rope", new UnityAction[1]
        {
            (UnityAction) (() => Utils.OpenFloatInputDialog<GrapplingGunBehaviour>(this.maxDistance * 1.0f, this, 
            (System.Action<GrapplingGunBehaviour, float>) ((t, v) => t.maxDistance = Mathf.Max(v, t.minDistance + 0.5f)), 
            "Set rope length in meters", " "))
        }));
        buttons.Add(new ContextMenuButton("resetLength", "Reset rope length", "Resets the rope length to default", () =>
        {
            this.maxDistance = 5000f;
        }));
    } 
    #endregion

    #region Runtime
    private void FixedUpdate()
    {
        if(!launched)
            return;
        if(winchState == GrapplingGunBehaviour.WinchStateEnum.Retracting)
        {
            RetractWinch();
        }else if(winchState == GrapplingGunBehaviour.WinchStateEnum.Extending)
        {   
            ExtendWinch();
        }
    }
    private void LateUpdate()
    {
        if(!winch) 
            return;
        lr.SetPosition(0, GetBarrelPosition());
        lr.SetPosition(1, hookRef.transform.TransformPoint(winch.connectedAnchor));
    }
    public void Use(ActivationPropagation activation)
    {
        if(!launched)
        {
            Launch();
        }else
        {
            Recover();
        }
    }
    #endregion

    #region Hook
    public void Launch() => StartCoroutine(LaunchRoutine());
    public void Recover() => StartCoroutine(RecoverRoutine());
    private IEnumerator LaunchRoutine()
    {
        yield return (object) new WaitForFixedUpdate();

        launched = true;
        
        foreach (UnityEngine.Object linkedObject in linkedObjects)
        {
            UnityEngine.Object.Destroy(linkedObject);
            UnityEngine.Object.Destroy(hookRef);
        }
        hookRef = (GameObject) null;

        Quaternion rotation = (double) transform.lossyScale.x < 0.0 ? Quaternion.Euler(0.0f, 0.0f, 180f) * transform.rotation : transform.rotation;
        PhysicalBehaviour hookPhys;
        hookRef = UnityEngine.Object.Instantiate<GameObject>(projectileAsset.Prefab, (Vector3) GetBarrelPosition(), rotation, transform.parent);
        sr.sprite = text[1];
        hookPhys = hookRef.GetComponent<PhysicalBehaviour>();
        hookPhys.SpawnSpawnParticles = false;
        hookRef.name = projectileAsset.name;
        CatalogBehaviour.PerformMod(projectileAsset, hookRef);
        hookRef.layer = -1;
        hookPhys.rigidbody.AddForce((Vector2) (hookRef.transform.right * projectileLaunchStrength), ForceMode2D.Impulse);
        rb.AddForce((Vector2) (-projectileLaunchStrength * recoilMultiplier * hookRef.transform.right), ForceMode2D.Impulse);
        linkedObjects.Add((UnityEngine.Object) hookRef);
        
        hookRef.GetComponent<HookBehaviour>().connectedGun = this;

        winch = gameObject.AddComponent<DistanceJoint2D>();
        winch.enabled = true;

        winch.autoConfigureDistance = false;
        winch.maxDistanceOnly = true;
        winch.distance = maxDistance;
        winch.connectedBody = hookRef.GetComponent<Rigidbody2D>();
        winch.connectedAnchor = new Vector2(-0.15f, 0.0f);
        winch.anchor = barrelPosition;

        lr.positionCount = 2;
    }
    private IEnumerator RecoverRoutine()
    {
        yield return (object) new WaitForFixedUpdate();

        foreach (UnityEngine.Object linkedObject in linkedObjects)
        {
            UnityEngine.Object.Destroy(linkedObject);
            UnityEngine.Object.Destroy(hookRef);
        }
        hookRef = (GameObject) null;

        sr.sprite = text[0];
        
        Destroy(winch);
        winch = (DistanceJoint2D) null;
        
        launched = false;
        lr.positionCount = 0;
        ResetWinchVars();   
    }
    #endregion
    
    #region Winch Functions
    public void UpdateDistance()
    {
        winch.distance = Vector2.Distance(transform.position, hookRef.transform.position);
    }
    private void ResetWinchVars()
    {
        winchState = GrapplingGunBehaviour.WinchStateEnum.Idle;
        StopAllCoroutines();
    }
    private void RetractWinch()
    {
        if(this.winch.distance > minDistance)
        {
            this.winch.distance -= 0.015f;
        }else
        {   
            this.ResetWinchVars();
        }
    }
    private void ExtendWinch()
    {
        if(this.winch.distance < maxDistance)
        {
            this.winch.distance += 0.045f;
        }else
        {
            this.ResetWinchVars();
        } 
    }
    #endregion

    #region Destroy
    private void OnDestroy()
    {
        foreach (UnityEngine.Object linkedObject in linkedObjects)
        {
            UnityEngine.Object.Destroy(linkedObject);
            UnityEngine.Object.Destroy(hookRef);
        }
        Recover();
    } 
    #endregion

    private enum WinchStateEnum
    {
        Idle,
        Retracting,
        Extending
    }
}