using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using TMPro;
using UnityEngine;
using UnityEngine.Events;

public class RaycastGrapplingGunBehaviour : MonoBehaviour, Messages.IUse
{   
    #region Objs
    private Rigidbody2D rb;   
    private PhysicalBehaviour phys;
    private LineRenderer lr;
    private RaycastHit2D hit;
    #endregion

    #region rope
    private DistanceJoint2D rope;
    #endregion

    #region Vectors
    public Vector2 barrelPosition;
    public Vector2 barrelDirection;
    public Vector2 GetBarrelPosition() => (Vector2) this.transform.TransformPoint((Vector3) this.barrelPosition);
    public Vector2 GetBarrelDirection() => (Vector2) (this.transform.TransformDirection((Vector3) this.barrelDirection) * this.transform.localScale.x);
    #endregion

    #region Vars
    private RopeStateEnum ropeState;
    private bool launched;
    private float maxDistance = 5000f;
    private float minDistance = 0.15f;
    #endregion 

    #region Setup
    private void Awake()
    {
        rb = GetComponent<Rigidbody2D>();
        phys = GetComponent<PhysicalBehaviour>();

        ropeState = RaycastGrapplingGunBehaviour.RopeStateEnum.Idle;
    } 
    private void Start(){
        lr = gameObject.AddComponent<LineRenderer>();

        lr.material = ModAPI.FindMaterial("Sprites-Default");
        lr.startColor = lr.endColor = Color.black;
        lr.startWidth = 0.05f;
        lr.numCapVertices = 10;
        lr.numCornerVertices = 10;
        lr.positionCount = 0;

        AddContextMenu();    
    }
    private void AddContextMenu()
    {

        List<ContextMenuButton> buttons = GetComponent<PhysicalBehaviour>().ContextMenuOptions.Buttons;
        buttons.Add(new ContextMenuButton("Stop rope", "Stop", "Stop the grappling gun's rope", () =>
        {
            ResetropeVars();
        }));
        buttons.Add(new ContextMenuButton("retractrope", "Retract", "Retract the grappling gun's rope", () =>
        {
            ResetropeVars();
            ropeState = RaycastGrapplingGunBehaviour.RopeStateEnum.Retracting;
        }));
        buttons.Add(new ContextMenuButton("extendrope", "Extend", "Extend the grappling gun's rope", () =>
        {
            ResetropeVars();
            ropeState = RaycastGrapplingGunBehaviour.RopeStateEnum.Extending;
        }));
        buttons.Add(new ContextMenuButton("setMaxLength", "Set rope length", "Set the length of the rope", new UnityAction[1]
        {
            (UnityAction) (() => Utils.OpenFloatInputDialog<RaycastGrapplingGunBehaviour>(this.maxDistance * 1.0f, this, 
            (System.Action<RaycastGrapplingGunBehaviour, float>) ((t, v) => t.maxDistance = Mathf.Max(v, t.minDistance + 0.5f)), 
            "Set rope length in meters", " "))
        }));
        buttons.Add(new ContextMenuButton("resetLength", "Reset rope length", "Resets the rope length to default", () =>
        {
            this.maxDistance = 5000f;
        }));
    } 
    #endregion

    #region Runtime
    private void FixedUpdate()
    {
        if(!launched)
            return;
        if(ropeState == RaycastGrapplingGunBehaviour.RopeStateEnum.Retracting)
        {
            Retractrope();
        }else if(ropeState == RaycastGrapplingGunBehaviour.RopeStateEnum.Extending)
        {   
            Extendrope();
        }
    }
    private void LateUpdate()
    {
        if(!rope) 
            return;
        lr.SetPosition(0, GetBarrelPosition());
        if(rope.connectedBody != null)
            lr.SetPosition(1, hit.transform.TransformPoint(rope.connectedAnchor));
        else
            lr.SetPosition(1, hit.point);
    }
    public void Use(ActivationPropagation activation)
    {
        if(!launched)
        {
            Shoot();
        }else
        {
            Recover();
        }
    }
    #endregion

    #region Hook
    private void Shoot()
    {
        hit = Physics2D.Raycast(transform.position, transform.right, maxDistance * transform.lossyScale.x, LayerMask.GetMask("Objects", "Bounds", "CollidingDebris"));

        if (hit.collider == null)
            return;

        launched = true;

        rope = gameObject.AddComponent<DistanceJoint2D>();
        rope.enabled = true;

        rope.autoConfigureDistance = false;
        rope.maxDistanceOnly = true;
        rope.anchor = barrelPosition;
        
        if(hit.transform.gameObject.layer == 9)
        {
            rope.connectedAnchor = hit.transform.InverseTransformPoint(hit.point);
            rope.connectedBody = hit.transform.GetComponent<Rigidbody2D>();
            
        }else
        {
            rope.connectedAnchor = hit.point;
            rope.connectedBody = null;
        }
        rope.distance = Vector2.Distance(transform.position, hit.point);
        lr.positionCount = 2;
    }
    private void Recover()
    {
        launched = false;

        Destroy(rope);
        rope = (DistanceJoint2D) null;

        lr.positionCount = 0;

        ResetropeVars();   
    }
    #endregion
    
    #region rope Functions
    private void ResetropeVars()
    {
        ropeState = RaycastGrapplingGunBehaviour.RopeStateEnum.Idle;
        StopAllCoroutines();
    }
    private void Retractrope()
    {
        if(this.rope.distance > minDistance)
        {
            this.rope.distance -= 0.015f;
        }else
        {   
            this.ResetropeVars();
        }
    }
    private void Extendrope()
    {
        if(this.rope.distance < maxDistance)
        {
            this.rope.distance += 0.045f;
        }else
        {
            this.ResetropeVars();
        } 
    }
    #endregion

    #region Destroy
    private void OnDestroy()
    {
        Recover();
    } 
    #endregion

    private enum RopeStateEnum
    {
        Idle,
        Retracting,
        Extending
    }
}
